/************************************************************************
 *									*
			Copyright (C) 1993 by
	Massachusetts Institute of Technology, Cambridge, MA
			 All rights reserved
 *									*
 ************************************************************************/

#include <malloc.h>
#include "gen.h"

/* Function: vectalloc()
 * Purpose: Allocate a vector structure
 * Method: The vector structure is defined in the header file "gen.h".
 *         It holds a single homogeneous coordinate with fields x,y,z,w.
 *         We use the function gen_array1 to allocate dynamic memory
 *         from the system heap.
 *         The vector is normalized with w = 1, in other words, x,y,z
 *         are a cartesian point.
 * Return: the address of the vector structure
 */
/* Functions referenced by vectalloc() are:
 *  errormsg()
 *  gen_array1()
 */
/* Functions that reference vectalloc() are:
 *  addpoints()
 *  add_vect()
 *  alloc_egeompts()
 *  alloc_fgeompts()
 *  approx_cubic_curv()
 *  approx_cubic_curv1()
 *  approx_cubic_surf()
 *  approx_user_knots()
 *  arc_length()
 *  boehm_curve()
 *  boehm_surface()
 *  CalcIsophotes()
 *  CalcReflectionLines()
 *  calc_knots_foley()
 *  calc_par_foley()
 *  calc_par_hartley()
 *  Camber2dTo3d()
 *  Camber2d_3d()
 *  CamberBisect()
 *  CamberBisect3d()
 *  camber_3d()
 *  camber_line_2d()
 *  camber_line_3d()
 *  camber_surf_3d()
 *  check_cv()
 *  check_patch_u()
 *  construct_bspl()
 *  construct_jbspl()
 *  ConvexHullSurf()
 *  convexhull_surf()
 *  convexhull_test()
 *  con_confun()
 *  copyvector()
 *  corner_offset()
 *  cross()
 *  evalbsp()
 *  evalbsp_per()
 *  evalderivbsp()
 *  evalderivbsp_per()
 *  evalderivsurf()
 *  evalderivsurf_per()
 *  evaldersurf()
 *  evaldersurf_per()
 *  evalsurf()
 *  evalsurf1()
 *  evalsurf1_per()
 *  EvalSurfAndCyl()
 *  EvalSurfAndCyl3d()
 *  evalsurf_per()
 *  eval_LE2D()
 *  eval_LE3D()
 *  extract_edge()
 *  fair_knot()
 *  fair_per_knot()
 *  fcn()
 *  find_dir()
 *  find_error_curv()
 *  find_error_curv1()
 *  find_error_surf1()
 *  fit_curve()
 *  fn_camber()
 *  fn_camber3()
 *  fn_thckns()
 *  fn_thckns3()
 *  footpoint_fcn()
 *  geod()
 *  geod_par()
 *  GetCurvValues()
 *  glue_vector()
 *  hodograph_surf()
 *  init_ortho_proj()
 *  IntersectBone()
 *  inters_stlines()
 *  localize_diagnostic()
 *  localize_sumsq()
 *  localize_sumsq_2d()
 *  localize_sumsq_opt()
 *  main()
 *  matr_m_vect()
 *  merge_tol_edges()
 *  midpoint()
 *  MinDistance()
 *  MinDistCurvCB()
 *  monotobern()
 *  offset_curve()
 *  opt_param()
 *  opt_param_per()
 *  OrthoDistance()
 *  output_2D()
 *  parallel()
 *  ParCurv_hodograph()
 *  putcontpts()
 *  raise_byone()
 *  SaveIgesCos()
 *  scale_vect()
 *  soslo()
 *  store_vertices()
 *  subbezier()
 *  SubDistance()
 *  subdivbs()
 *  sub_vect()
 *  torus_gen()
 *  transformegeom()
 *  transformfgeom()
 *  TransformGridSurf()
 *  TransformListCurv()
 *  unitvector()
 *  VectorBox()
 *  vector_dircos()
 *  vector_rotate()
 */

vector *vectalloc(void)		/* allocate the memory for a vector */
{
  vector *v;
  v = (vector *) gen_array1 (1, sizeof(vector));
  if (v)
    v->w = 1.0;
  else
    errormsg (0, "allocation failure in vectalloc()");

  return (v);
}

/* Function: vectfree()
 * Purpose: Deallocate vector structure
 * Method: The function free_garray1 is called to deallocate the 
 *         vector structure.  The input address is not checked to
 *         see if it is valid.
 * Arguments:
 *  v - the address of a previously allocated vector stucture
 */
/* Functions referenced by vectfree() are:
 *  free_garray1()
 */
/* Functions that reference vectfree() are:
 *  approx_user_knots()
 *  arc_length()
 *  bias()
 *  Bisector()
 *  bone_intersect()
 *  bone_intersect3()
 *  build_offsets_int()
 *  CalcHubTip()
 *  calc_disc()
 *  calc_points_curv()
 *  calc_points_surf()
 *  Camber2dTo3d()
 *  Camber2d_3d()
 *  CamberBisect()
 *  CamberBisect3d()
 *  camber_3d()
 *  camber_line_2d()
 *  camber_line_3d()
 *  camber_surface()
 *  camber_surf_3d()
 *  check_cv()
 *  check_patch_u()
 *  compute_diffeq()
 *  compute_ortho_proj()
 *  construct_bspl()
 *  construct_jbspl()
 *  ContourFacets()
 *  con_confun()
 *  corner_offset()
 *  cross_link()
 *  CurveOrientation()
 *  CylinderTo3D()
 *  cyl_frenet_tr()
 *  cyl_generatrix()
 *  cyl_sample_gencyl()
 *  DrawCosKnots()
 *  DrawCurvKnots()
 *  DrawFacetEvaluateMap()
 *  DrawSurfKnots()
 *  DrawSurfSplits()
 *  DrawTrimWireframe()
 *  EvalMaxCamber()
 *  EvalMaxCamber3d()
 *  EvalSurfAndCyl()
 *  EvalSurfAndCyl3d()
 *  EvaluateCos()
 *  EvaluateCurv()
 *  EvaluateSurf()
 *  eval_curve_bounded()
 *  eval_cylsect_param()
 *  eval_cylsect_param3()
 *  eval_vert()
 *  fair_per_knot()
 *  fcn_crit()
 *  fcn_crit3()
 *  fdist_surf1()
 *  FilterNegative()
 *  find_error_surf()
 *  fn_camber3()
 *  footpoint_fcn()
 *  free_varray1()
 *  free_varray2()
 *  frenet_tr()
 *  FuncSurfAndCyl()
 *  FuncSurfAndCyl3d()
 *  functcv_sqr()
 *  fun_camber()
 *  fun_camber3()
 *  generatrix()
 *  GetCosValues()
 *  GetCurvValues()
 *  GetSurfValues()
 *  g_camber()
 *  g_camber3()
 *  init_ortho_proj()
 *  init_proj()
 *  integral_build_offset()
 *  integral_sample_offset()
 *  IntersectBone()
 *  int_sample_gencyl()
 *  IsoLoopIntersect()
 *  iso_curv0()
 *  iso_curv1()
 *  lead_chord()
 *  LinearCosNormal()
 *  localize_diagnostic()
 *  localize_sumsq()
 *  localize_sumsq_2d()
 *  localize_sumsq_opt()
 *  loft_integral()
 *  MakeCamberLine3d()
 *  max_chord()
 *  max_cmbr()
 *  max_cmbr3()
 *  max_curve_curvature()
 *  merge_tol_edges()
 *  next_u()
 *  normalcurv()
 *  normaldu()
 *  offnorm2d()
 *  opt_param()
 *  opt_param_per()
 *  ParCurv_to_PowCurv()
 *  ParSurf_to_PowSurf()
 *  partl_u_functcv_sqr()
 *  point_bspl()
 *  position_err()
 *  PostScriptCosKnots()
 *  PostScriptCurvKnots()
 *  PostScriptFacetEvaluateMap()
 *  PostScriptSurfKnots()
 *  PowParCurv_compare()
 *  PowParSurf_compare()
 *  PrimitiveCurvCB()
 *  rbspcur()
 *  revalderivsurf()
 *  revalderivsurf_per()
 *  RobustCurvCB()
 *  RobustDistCB()
 *  SampleCurvCB()
 *  SampleSurfGridCB()
 *  SampleSurfListCB()
 *  sample_blend()
 *  sample_gencyl()
 *  sample_mod_surf()
 *  scattered_fit()
 *  SpanwiseU()
 *  StoreCosValues()
 *  StoreCurvValues()
 *  StoreSurfValues()
 *  StoreTrimValues()
 *  SubDistance()
 *  surf_data()
 *  test_tangent()
 *  tol_sample_offsets()
 *  TransformGridSurf()
 *  TransformListCurv()
 *  TrimPoints()
 *  TrimSection()
 *  TrimSection3d()
 *  trim_section()
 *  trim_section3()
 *  trim_start_par()
 *  trim_start_par3()
 *  vectang()
 *  vec_rot()
 *  vec_rot_fix()
 *  WriteIges112()
 */

void vectfree(vector *v)
{
  free_garray1((char *)v);
}
